use std::thread;
use std::sync::Arc;
use std::sync::Mutex;
use std::sync::mpsc;

fn move_borrow() {
    // // cannot borrow v:
    // // comportement possible: le programme a fini avant qu'on ait pu faire un push par exemple.
    // let mut v = Vec::new();
    // let loc = thread::spawn(|| {
    //     v.push(1);
    // });
    // v.push(4);
    // println!("{:?}", loc.join().unwrap());

    // // v moved in thread, cannot be used elsewhere
    // let mut v = Vec::new();
    // let loc = thread::spawn(move || {
    //     v.push(1);
    //     v
    // });
    // v.push(4);
    // println!("{:?}", loc.join().unwrap());

    // here it works
    let mut v = Vec::new();
    let loc = thread::spawn(move || {
        v.push(1);
        v
    });

    println!("{:?}", loc.join().unwrap());
}

fn share_with_rc() {
    // let v = Rc::new(vec![1,2,3,4]); // reference counted
    // let v_cloned = v.clone(); // create a new reference
    
    // // rc cannot be sent between threads safely.
    // let handle = thread::spawn(move || {
    //     println!("{:?}", v);
    // });
    // handle.join();

    // Arc can be sent between threads safely.
    let v = Arc::new(vec![1,2,3,4]); // reference counted
    let v_cloned = v.clone(); // create a new reference
    let handle = thread::spawn(move || {
        println!("{:?}", v);
    });

    handle.join().unwrap();

    println!("{:?}", v_cloned);

}

fn mutate_with_mutex() {
    // Arc can be sent between threads safely.
    let v = Arc::new(Mutex::new(vec![1,2,3,4])); // reference counted
    let v_cloned = v.clone(); // create a new reference
    let handle = thread::spawn(move || {
        let mut v = v_cloned.lock().unwrap();

        v.push(5);
    });

    handle.join().unwrap();

    println!("{:?}", v);

}

fn communicate_with_channel() {
    let (tx, rx) = mpsc::channel();

    let tx_cloned = tx.clone(); // rx cannot be cloned

    let hdl = thread::spawn(move || {
        tx.send(vec![1,2,3]).unwrap();
    });

    let hdl2 = thread::spawn(move || {
        tx_cloned.send(vec![4,5,6]).unwrap();
    });

    println!("Recieved {:?}", rx.recv());
    println!("Recieved {:?}", rx.recv());

    hdl.join().unwrap();
    hdl2.join().unwrap();
}



fn spawn_unordered_not_finishing() {
    for i in 0..10 {
        let _loc = thread::spawn(move || {
            println!("Hello from child thread {}!", i);
        });
    }

    for i in 0..10 {
        println!("Hello {} from main thread!", i);
    }
}

fn spawn_unordered_join() {
    let mut handles = Vec::new();
    for i in 0..10 {
        handles.push(thread::spawn(move || {
            println!("Hello from child thread {}!", i);
        }));
    }

    for h in handles {
        h.join().unwrap();
    }
}

fn hello_world() {
    let loc = thread::spawn(|| {
        "World"
    });

    println!("Hello {}!", loc.join().unwrap());
}

fn main() {
    hello_world();

    spawn_unordered_not_finishing();
    
    spawn_unordered_join();

    move_borrow();

    share_with_rc();

    mutate_with_mutex();

    communicate_with_channel();
}
